# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import enum
import typing

import attrs

from .base import Component

if typing.TYPE_CHECKING:
    from wyvern.models.channels import ChannelType
    from wyvern.models.emojis import CustomEmoji

__all__: tuple[str, ...] = ("SelectType", "Select", "SelectOption")


class SelectType(enum.IntEnum):
    """Type of the select option"""

    STRING = 3
    """String type select, this requires the `options` kwarg in the [Select][] class."""
    USER = 5
    """User type select menu."""
    ROLE = 6
    """Role type select menu."""
    MENTIONABLE = 7
    """Mentionable type select menu."""
    CHANNEL = 8
    """Channel type select menu."""


@attrs.define(kw_only=True, slots=True)
class SelectOption:
    """Represents a Select option."""

    label: str
    """Label of the option."""
    _value: str | None = None
    """Value of the option, defaults to the label."""
    description: str | None
    """Description of the option."""
    emoji: str | CustomEmoji | None = None
    """Emoji attached to the option."""
    default: bool = False
    """Set to `True` if the option is to be used as default."""

    @property
    def value(self) -> str:
        return self._value or self.label

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "label": self.label,
            "value": self.value,
            "description": self.description,
            "emoji": str(self.emoji),
            "default": self.default,
        }


@attrs.define(kw_only=True, slots=True)
class Select(Component):
    """Represents a discord Select menu."""

    type: SelectType = SelectType.STRING
    """Type of the select"""
    custom_id: str
    """Custom ID of the select."""
    options: list[SelectOption] = []
    """Options attached to the select."""
    channel_types: list[ChannelType] | None = None
    """Channel types for CHANNEL type select."""
    min_values: int | None = None
    """Minimum selections."""
    max_values: int | None = None
    """Maximum selections."""
    disabled: bool = False
    """True if the select is disabled."""

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "type": self.type,
            "custom_id": self.custom_id,
            "options": [opt.to_payload() for opt in self.options],
            "channel_types": self.channel_types,
            "min_values": self.max_values,
            "max_values": self.max_values,
            "disabled": self.disabled,
        }



import argparse
import sys

import colorama

from wyvern import __version__

parser = argparse.ArgumentParser(
    prog="wyvern", description="Commands related to the library and it's usage.", add_help=True
)
parser.add_argument("-V", "-v", "--version", help="Check the version of library.", action="store_true")


def main() -> None:
    args = parser.parse_args()
    if args.version is True:
        print(colorama.Fore.LIGHTCYAN_EX, end="")
        print(
            "Library version:",
            colorama.Style.BRIGHT,
            __version__,
            colorama.Style.NORMAL,
            "\nPython version:",
            colorama.Style.BRIGHT,
            sys.version,
            colorama.Style.RESET_ALL,
        )


main()



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from wyvern._internals import _ListenerArg
    from wyvern.clients import CommandsClient
    from wyvern.commands.slash_commands import CommandChoice
    from wyvern.components.buttons import Button
    from wyvern.interactions.applications import (
        ApplicationCommandInteraction,
        AutocompleteInteraction,
        InteractionOption,
    )
    from wyvern.interactions.components import ComponentInteraction

    CheckT: TypeAlias = typing.Callable[..., typing.Awaitable[bool]]
    AnyCallableT: TypeAlias = typing.Callable[..., typing.Awaitable[typing.Any]]
    CallbackT: TypeAlias = typing.Callable[..., typing.Awaitable[typing.Any]]
    AutocompleteT: TypeAlias = typing.Callable[
        [AutocompleteInteraction, InteractionOption],
        typing.Awaitable[typing.Sequence[typing.Union[CommandChoice, str]]],
    ]
    ButtonCallbackT: TypeAlias = typing.Callable[[Button, ComponentInteraction], typing.Awaitable[typing.Any]]

    class AppCommandCallbackT(typing.Protocol):
        @typing.overload
        def __call__(
            self, inter_or_client: CommandsClient, inter: ApplicationCommandInteraction, /, *args: typing.Any
        ) -> typing.Any:
            ...

        @typing.overload
        def __call__(self, inter_or_client: ApplicationCommandInteraction, /, *args: typing.Any) -> typing.Any:
            ...

        def __call__(
            self, inter_or_client: ApplicationCommandInteraction | CommandsClient, /, *args: typing.Any
        ) -> typing.Any:
            ...

    class EventListenerCallbackT(typing.Protocol):
        def __call__(self, *args: _ListenerArg) -> typing.Any:
            ...



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import typing

import attrs

__all__: tuple[str, ...] = ("Task", "task")


@attrs.define(kw_only=True, slots=True)
class Task:
    """Represents a task that gets triggerd after some
    interval of time repeatedly."""

    trigger: typing.Callable[..., typing.Awaitable[typing.Any]]
    """The coro to trigger at every interval."""
    delay: float
    """Time delay between triggers ( in seconds )"""
    wait_until_complete: bool = True
    """Weather to wait before one trigger is complete."""
    is_running: bool = False
    """True if the task is running."""

    def update_delay(self, delay: float) -> None:
        self.delay = delay

    async def _runner(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        loop = asyncio.get_event_loop()
        while self.is_running is True:
            if self.wait_until_complete is False:
                loop.create_task(self.trigger(*args, **kwargs))  # type: ignore
            else:
                await self.trigger(*args, **kwargs)
            await asyncio.sleep(self.delay)

    def run(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Starts the task."""
        if self.is_running is False:
            self.is_running = True
        else:
            raise Exception("The task is already running.")
        loop = asyncio.get_event_loop()
        loop.create_task(self._runner(*args, *kwargs))

    def stop(self) -> None:
        """Stops the task."""
        self.is_running = False


def task(
    s: float | None = None, m: float | None = None, h: float | None = None, wait_until_complete: bool = True
) -> typing.Callable[..., Task]:
    """Interface to create a task.

    Parameters
    ----------

    WIP : typing.Any
        Docs to be added.
    """

    def inner(trigger: typing.Callable[..., typing.Awaitable[typing.Any]]) -> Task:
        nonlocal s, m, h
        if (delays := len([item for item in [s, m, h] if item is not None])) > 1 or delays == 0:
            raise ValueError("Only one delay field can be used for the decorator.")
        delay = 0.0
        if s:
            delay = s
        elif m:
            delay = m * 60.0
        elif h:
            delay = h * 60 * 60.0

        return Task(
            delay=delay,
            trigger=trigger,
            wait_until_complete=wait_until_complete,
        )

    return inner



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Slash commands support."""

from __future__ import annotations

import typing

import attrs

from wyvern import interactions, utils
from wyvern.interactions.base import InteractionCommandOptionType as OptionType

from .base import BaseCallable, HasAutoComplete, HasSubCommands, SlashOptionConverter

if typing.TYPE_CHECKING:
    from wyvern._types import AppCommandCallbackT
    from wyvern.clients import CommandsClient
    from wyvern.interactions import ApplicationCommandInteraction, Localizations
    from wyvern.models.channels import ChannelType

__all__: tuple[str, ...] = (
    "CommandChoice",
    "OptionType",
    "as_slash_command",
    "with_option",
    "SlashCommand",
    "SlashGroup",
)


@attrs.define(kw_only=True)
class CommandChoice:
    """Represents a slash command choice.

    Parameters
    ----------

    name : str
        Name to show on discord.
    value : str | int
        Value of the choice.

    """

    name: str
    value: str | int
    name_locales: dict[str, Localizations] | utils.Empty = utils.EMPTY

    def to_payload(self) -> dict[str, str | int | dict[str, dict[str, str]]]:
        return {
            "name": self.name,
            "value": self.value or self.name,
            "name_localizations": {name: local.final_data for name, local in _locale.items()}
            if not isinstance((_locale := self.name_locales), utils.Empty)
            else {},
        }


@attrs.define(kw_only=True)
class CommandOption:
    """Represents a slash command option. (In the constructor)"""

    name: str
    """Name of the option."""
    name_locales: Localizations | utils.Empty = utils.EMPTY
    """Name locales for the option."""
    description: str
    """Description of the option."""
    description_locales: Localizations | utils.Empty = utils.EMPTY
    """Description locales for the option."""
    default: typing.Any = utils.EMPTY
    """Default value for the option"""
    type: OptionType
    """Type of the option."""
    converter: SlashOptionConverter | None = None
    """A custom converter for the option, if any."""
    autocomplete: bool = False
    """Weather to enable autocomplete for this option or not."""
    choices: list[CommandChoice | str | int] = []
    """List of choices to add to the option."""
    channel_types: list[ChannelType] = []
    """List of channel types for channel options."""
    min_value: int | None = None
    """Minimum value for the option"""
    max_value: int | None = None
    """Maximum value for the option"""
    min_length: int | None = None
    """Minimum length acceptable for the option"""
    max_length: int | None = None
    """Maximum length acceptable for the option"""

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
            "required": False if self.default is utils.EMPTY else True,
            "type": self.type,
            "autocomplete": self.autocomplete,
            "choices": [
                (choice.to_payload() if isinstance(choice, CommandChoice) else {"name": choice, "value": choice})
                for choice in self.choices
            ],
            "channel_types": self.channel_types,
            "min_value": self.min_value,
            "min_length": self.min_length,
            "max_length": self.max_length,
            "max_value": self.max_value,
            "name_localizations": _locale.final_data
            if not isinstance((_locale := self.name_locales), utils.Empty)
            else {},
            "description_localizations": _locale.final_data
            if not isinstance((_locale := self.description_locales), utils.Empty)
            else {},
        }


@attrs.define(kw_only=True, slots=True, repr=True)
class SlashCommand(BaseCallable, HasAutoComplete):
    """Represents a slash command constructor."""

    name: str
    """Name of the command."""
    description: str
    """Description of the command."""
    options: list[CommandOption] = attrs.field(init=False, default=[])
    """List of command options."""
    guild_only: bool = False
    """Set to True for command to appear in guilds only."""
    client: CommandsClient | None = attrs.field(init=False, default=None)
    callback: AppCommandCallbackT
    guild_ids: list[int] = []
    """"""

    def __str__(self) -> str:
        return self.name

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        await self.callback(*args, **kwargs)

    def _set_client(self, client: CommandsClient) -> "SlashCommand":
        self.client = client
        return self

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
            "options": [option.to_payload() for option in self.options],
            "dm_permissions": not self.guild_only,
            "type": interactions.base.InteractionCommandType.MESSAGE,
        }

    async def with_autocomplete(
        self, opt_name: str | None = None, *, opt_names: typing.Sequence[str] = ()
    ) -> typing.Callable[..., typing.Callable[..., typing.Awaitable[typing.Sequence[CommandChoice | str]]]]:
        """Interface for creating autocomplete callabcks.

        Parameters
        ----------

        opt_name : str
            Name of the option this autocomplete is for.
        opt_names : typing.Sequence[str]
            Sequence of options this autocomplete is for.

        !!! note
            opt_names overrides opt_name

        ??? example
            ```py

            from __future__ import annotations

            import wyvern
            from wyvern import commands

            client = wyvern.CommandsClient("TOKEN")


            @commands.with_option(
                name="color", description="name of color to show hex of.", type=commands.OptionType.INTEGER, autocomplete=True
            )
            @client.slash_command(name="color_hex", description="gives you color hexs")
            async def colors(inter: wyvern.ApplicationCommandInteraction, color: int) -> None:
                await inter.create_message_response(f"Hex for the color was {hex}")


            @colors.with_autocomplete(opt_name="color")
            async def autocomplete_inter(
                inter: wyvern.AutocompleteInteraction, option: wyvern.InteractionOption
            ) -> list[commands.CommandChoice]:
                colors = [commands.CommandChoice(name="blue", value=0x4F62A3), commands.CommandChoice(name="red", value=0xD11417)]
                return [color for color in colors if color.name in inter.data.value]


            client.run()

            ```

        """  # noqa: E501
        return await super().with_autocomplete(opt_name, opt_names=opt_names)

    async def create_command(self) -> SlashCommand:
        """Creates the command on discord ( if already doesnt exist ).

        Returns
        -------

        wyvern.commands.SlashCommand
            This same object
        """
        assert self.client
        await self.client.rest._create_app_command_from_payload(payload=self.to_payload())
        return self


@attrs.define(kw_only=True)
class SlashSubCommand(BaseCallable, HasAutoComplete):
    name: str
    description: str
    callback: AppCommandCallbackT

    options: list[CommandOption] = attrs.field(init=False, default=[])

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        await self.callback(*args, **kwargs)

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
            "options": [option.to_payload() for option in self.options],
        }


@attrs.define(kw_only=True)
class SlashSubGroupCommand:
    name: str
    description: str
    callback: AppCommandCallbackT

    options: list[CommandOption] = attrs.field(init=False, default=[])

    async def __call__(self, inter: ApplicationCommandInteraction, **kwargs: typing.Any) -> typing.Any:
        await self.callback(inter, **kwargs)

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
            "options": [option.to_payload() for option in self.options],
        }


@attrs.define(kw_only=True)
class SlashSubGroup(HasSubCommands):
    name: str
    description: str
    subcommands: dict[str, SlashSubGroupCommand] = attrs.field(init=False, default={})

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "type": OptionType.SUB_COMMAND_GROUP,
            "name": self.name,
            "description": self.description,
            "options": [opt.to_payload() for opt in self.subcommands.values()],
        }

    def with_subcommand(
        self, *, name: str, description: str
    ) -> typing.Callable[[AppCommandCallbackT], SlashSubGroupCommand]:
        def inner(callback: AppCommandCallbackT) -> SlashSubGroupCommand:
            self.subcommands[name] = (
                subcmd := SlashSubGroupCommand(name=name, description=description, callback=callback)
            )
            return subcmd

        return inner


@attrs.define(kw_only=True)
class SlashGroup(HasSubCommands):
    name: str
    description: str
    guild_ids: list[int]
    guild_only: bool = False
    subcommands: dict[str, SlashSubCommand] = attrs.field(init=False, default={})
    subgroups: dict[str, SlashSubGroup] = attrs.field(init=False, default={})
    client: "CommandsClient" | None = attrs.field(init=False, default=None)

    def _set_client(self, client: "CommandsClient") -> "SlashGroup":
        self._client = client
        return self

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
            "options": self._options,
            "dm_permissions": not self.guild_only,
            "type": interactions.base.InteractionCommandType.MESSAGE,
        }

    @property
    def _options(self) -> list[dict[str, typing.Any]]:
        opts = list(self.subcommands.values()) + list(self.subgroups.values())
        return [opt.to_payload() for opt in opts]

    def with_subgroup(self, *, name: str, description: str) -> SlashSubGroup:
        self.subgroups[name] = (group := SlashSubGroup(name=name, description=description))
        return group

    def with_subcommand(
        self, *, name: str, description: str
    ) -> typing.Callable[[AppCommandCallbackT], SlashSubCommand]:
        def inner(callback: AppCommandCallbackT) -> SlashSubCommand:
            nonlocal name, description
            self.subcommands[name] = (cmd := SlashSubCommand(name=name, description=description, callback=callback))
            return cmd

        return inner


def as_slash_command(
    *,
    name: str,
    description: str,
    guild_ids: typing.Sequence[int] = (),
    guild_only: bool = False,
) -> typing.Callable[[AppCommandCallbackT], SlashCommand]:
    """Creates a slash command object.

    Parameters
    ----------

    name : str
        Name of the command.
    description : str
        Description of the command.

    Returns
    -------

    typing.Callable[[AppCommandCallbackT], SlashCommand]

        A [SlashCommand][] when called.


    """

    def inner(callback: AppCommandCallbackT) -> SlashCommand:
        return SlashCommand(
            name=name, description=description, callback=callback, guild_ids=list(guild_ids), guild_only=guild_only
        )

    return inner


def as_slash_group(
    *, name: str, description: str, guild_ids: typing.Sequence[int] = (), guild_only: bool = False
) -> SlashGroup:
    return SlashGroup(name=name, description=description, guild_ids=list(guild_ids), guild_only=guild_only)


def as_option(
    *,
    name: str,
    description: str,
    type: OptionType = OptionType.STRING,
    default: typing.Any = utils.EMPTY,
    converter: SlashOptionConverter | None = None,
    autocomplete: bool = False,
    choices: typing.Sequence[CommandChoice | str | int] = [],
    channel_types: typing.Sequence["ChannelType"] = [],
    min_value: int | None = None,
    max_value: int | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    name_locales: Localizations | utils.Empty = utils.EMPTY,
    description_locales: Localizations | utils.Empty = utils.EMPTY,
) -> CommandOption:
    return CommandOption(
        name=name,
        description=description,
        default=default,
        type=type,
        converter=converter,
        autocomplete=autocomplete,
        choices=list(choices),
        channel_types=list(channel_types),
        min_length=min_length,
        min_value=min_value,
        max_length=max_length,
        max_value=max_value,
        name_locales=name_locales,
        description_locales=description_locales,
    )


def with_option(
    *,
    name: str,
    description: str,
    type: OptionType = OptionType.STRING,
    default: typing.Any = utils.EMPTY,
    converter: SlashOptionConverter | None = None,
    autocomplete: bool = False,
    choices: typing.Sequence[CommandChoice | str | int] = [],
    channel_types: typing.Sequence["ChannelType"] = [],
    min_value: int | None = None,
    max_value: int | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    name_locales: Localizations | utils.Empty = utils.EMPTY,
    description_locales: Localizations | utils.Empty = utils.EMPTY,
) -> typing.Callable[[SlashCommand], SlashCommand]:
    """Adds an option to a slash command.

    Parameters
    ----------

    name : str
        Name of the option.
    description : str
        Description of the option.
    type : OptionType
        Type of the option.
    default : typing.Any
        Default value for the option.
    convertor : SlashOptionConvertor | None
        A custom convertor if any.
    autcomplete : bool
        Is autcomplete enabled for this option ( defaults to False )

    Returns
    -------

    typing.Callable[[SlashCommand], SlashCommand]

        The command this option was added to.
    """

    def inner(command: SlashCommand) -> SlashCommand:
        command.options.append(
            CommandOption(
                name=name,
                description=description,
                default=default,
                type=type,
                converter=converter,
                autocomplete=autocomplete,
                choices=list(choices),
                channel_types=list(channel_types),
                min_length=min_length,
                min_value=min_value,
                max_length=max_length,
                max_value=max_value,
                name_locales=name_locales,
                description_locales=description_locales,
            )
        )
        return command

    return inner



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import dataclasses
import datetime
import typing

import aiohttp
import multidict

from wyvern import commands, interactions, models
from wyvern.exceptions import HTTPException

from .endpoints import Endpoints

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.components import ActionRowContainer, Modal
    from wyvern.constructors.embeds import EmbedConstructor

__all__: tuple[str, ...] = ("RESTClient",)


@dataclasses.dataclass
class RequestRoute:
    _url: str
    api_version: int = 10
    type: str = "GET"
    json: dict[str, typing.Any] | None = None

    @property
    def url(self) -> str:
        return f"https://discord.com/api/v{self.api_version}/{self._url}"


class RESTClient:
    """The REST Client that deals with disocrd REST Api requests."""

    def __init__(
        self,
        *,
        client: "GatewayClient",
        token: str,
        api_version: int = 10,
        client_session: aiohttp.ClientSession | None = None,
    ) -> None:
        self._client = client
        self._session: aiohttp.ClientSession
        self._token = token
        self._api_version = api_version
        self._headers: dict[str, multidict.istr] = {"Authorization": multidict.istr(f"Bot {token}")}

        if client_session is not None:
            self._session = client_session

    async def _create_websocket(self) -> aiohttp.ClientWebSocketResponse:
        if getattr(self, "_session", None) is None:
            self._session = aiohttp.ClientSession(headers=self._headers)
        return await self._session.ws_connect(  # type: ignore
            f"wss://gateway.discord.gg/?v={self._api_version}&encoding=json"
        )

    async def request(self, route: RequestRoute) -> typing.Any:
        headers = self._headers.copy()
        headers["Content-Type"] = multidict.istr("application/json")
        self._client._logger.debug(f"Creating a {route.type} request to {route._url} endpoint.")
        res = await self._session.request(route.type, route.url, headers=headers, json=route.json)
        if res.status in (200, 201):
            return await res.json()
        if res.status in (204, 304):
            return
        else:
            self.handle_error(exc := HTTPException((await res.json())["message"], res.status, route).create())
            raise exc

    def handle_error(self, exc: HTTPException) -> None:
        self._client._logger.error(
            "Exception while creating a HTTP request.\nType: %s, Endpoint: %s\nException: %s ",
            exc.route.type,
            exc.route.url,
            exc.message,
        )

    async def fetch_user(self, user_id: int) -> models.User:
        """Fetchs a user using the REST api.

        Parameters
        ----------

        user_id : int
            ID of the user that is to be fetched.

        Returns
        -------

        wyvern.models.users.User
            The user object that was fetched.

        Raises
        ------

        wyvern.exceptions.NotFound
            The targetted user was not found.
        """
        res = await self.request(RequestRoute(Endpoints.get_user(user_id)))
        return models._converters.payload_to_user(self._client, res)

    async def fetch_client_user(self) -> "models.BotUser":
        """
        Fetchs the bot's user object.

        Returns
        -------

        wyvern.models.users.BotUser
            BotUser object representating the bot's user.
        """
        res = await self.request(RequestRoute(Endpoints.get_current_user()))
        return models._converters.payload_to_botuser(self._client, res)

    async def edit_client_user(self, username: str | None = None, avatar: bytes | None = None) -> "models.BotUser":
        """Edits the bot's user.

        Parameters
        ----------

        username : str
            The new username.
        avatar : bytes
            The new avatar bytes.

        Returns
        -------

        wyvern.models.users.BotUser
            The updated user of bot.
        """

        payload: dict[str, bytes | str] = {}
        if username is not None:
            payload["username"] = username
        if avatar is not None:
            payload["avatar"] = avatar
        res: dict[str, int | str | bool] = await self.request(
            RequestRoute(Endpoints.get_current_user(), type="PATCH", json=payload)
        )
        return models._converters.payload_to_botuser(self._client, res)

    async def fetch_member(self, guild_id: int, member_id: int) -> models.Member:

        res = await self.request(
            RequestRoute(
                Endpoints.get_guild_member(guild_id, member_id),
            )
        )
        return models._converters.payload_to_member(self._client, models.Snowflake.create(guild_id), res)

    async def create_message(
        self,
        channel_id: int,
        content: str | None = None,
        *,
        embeds: typing.Sequence["EmbedConstructor"] = (),
        components: typing.Sequence[ActionRowContainer] = (),
        reference: int | models.MessageReference | None = None,
        allowed_mentions: models.AllowedMentions | None = None,
        flags: models.messages.MessageFlags | None = None,
        delete_after: int | None = None,
    ) -> "models.messages.Message":
        """Create a new message.

        Parameters
        ----------

        channel_id : int
            ID of the channel where the message is to be sent.
        content : str | None
            The text content of the message.
        embeds : typing.Sequence[wyvern.constructors.embeds.EmbedConstructor]
            Sequence of embeds to send.
        components : typing.Sequence[wyvern.components.container.ActionRowContainer]
            Sequence of action rows to send.
        reference : int | wyvern.models.messages.MessageReference | None
            ID or a message reference to which this is a response to.
        allowed_mentions : wyvern.models.messages.AllowedMentions | None
            Allowed mentions configs.

        Returns
        -------

        wyvern.models.messages.Message
            The message object that got created.

        """
        payload: dict[str, typing.Any] = {
            "content": content,
            "embeds": [embed._payload for embed in embeds],
            "components": [comp.to_payload() for comp in components],
            "allowed_mentions": (allowed_mentions or self._client.allowed_mentions).to_payload(),
            "flags": flags.value if flags else None,
        }

        if reference is not None:
            if isinstance(reference, models.MessageReference):
                payload["message_reference"] = reference.to_payload()
            else:
                payload["message_reference"] = models.MessageReference(message_id=reference).to_payload()

        res: dict[str, typing.Any] = await self.request(
            RequestRoute(Endpoints.create_message(channel_id), type="POST", json=payload),
        )

        msg = models._converters.payload_to_message(self._client, res)

        async def _delete_after(sec: int) -> None:

            await asyncio.sleep(sec)
            await self.delete_message(msg.channel_id, msg.id)

        if delete_after is not None:
            asyncio.create_task(_delete_after(delete_after))

        return msg

    async def delete_message(self, channel_id: int, message_id: int) -> None:
        await self.request(RequestRoute(Endpoints.delete_message(channel_id, message_id), type="DELETE"))

    async def fetch_message(self, channel_id: int, message_id: int) -> models.Message:

        res = await self.request(RequestRoute(Endpoints.get_channel_message(channel_id, message_id)))
        return models._converters.payload_to_message(self._client, res)

    async def fetch_messages(
        self, channel_id: int, around: datetime.datetime
    ) -> dict[models.Snowflake, models.Message]:
        try:
            res = await self.request(RequestRoute(Endpoints.get_channel_messages(channel_id)))
            return {
                msg.id: msg for msg in [models._converters.payload_to_message(self._client, payload) for payload in res]
            }
        except HTTPException as e:
            raise e

    async def create_application_command(
        self,
        *,
        name: str,
        description: str,
        options: typing.Sequence[commands.slash_commands.CommandOption] = (),
        dm_permission: bool = True,
        type: interactions.base.InteractionCommandType,
    ) -> typing.Any:
        payload = {
            "name": name,
            "description": description,
            "options": [option.to_payload() for option in options],
            "dm_permission": dm_permission,
            "type": type,
        }
        return await self.request(
            RequestRoute(Endpoints.interaction_command(self._client._client_id), type="POST", json=payload)
        )

    async def _create_app_command_from_payload(self, payload: dict[str, typing.Any]) -> typing.Any:
        return await self.request(
            RequestRoute(Endpoints.interaction_command(self._client._client_id), type="POST", json=payload)
        )

    async def create_interaction_response(
        self,
        interaction: interactions.Interaction,
        response_type: interactions.InteractionResponseType,
        *,
        content: str | None = None,
        embeds: typing.Sequence[EmbedConstructor] = (),
        components: typing.Sequence[ActionRowContainer] = (),
        flags: models.messages.MessageFlags | None = None,
        allowed_mentions: models.messages.AllowedMentions | None = None,
        modal: Modal | None = None,
    ) -> None:
        payload: dict[str, typing.Any] = {"type": int(response_type), "data": {}}
        if response_type is interactions.InteractionResponseType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE:
            pass
        elif response_type is interactions.InteractionResponseType.CHANNEL_MESSAGE_WITH_SOURCE:
            payload["data"]["content"] = content
            payload["data"]["embeds"] = [embed._payload for embed in embeds]
            payload["data"]["components"] = [builder.to_payload() for builder in components]
            payload["data"]["allowed_mentions"] = (allowed_mentions or self._client.allowed_mentions).to_payload()
            payload["data"]["flags"] = flags.value if flags else None
        elif response_type is interactions.InteractionResponseType.MODAL:
            if modal is None:
                raise ValueError("No Modal instance provided to send.")
            payload["data"]["custom_id"] = modal.custom_id
            payload["data"]["title"] = modal.title
            payload["data"]["components"] = [{"type": 1}]
            payload["data"]["components"][0]["components"] = [comp.to_payload() for comp in modal.text_inputs]
        await self.request(
            RequestRoute(
                Endpoints.interaction_callback(
                    interaction.id,
                    interaction.token,
                ),
                type="POST",
                json=payload,
            )
        )



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

__all__: tuple[str, ...] = ("Endpoints",)


class Endpoints:
    @classmethod
    def guild_audit_logs(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/audit-logs"

    @classmethod
    def list_auto_moderation_rules(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/auto-moderation/rules"

    @classmethod
    def get_auto_moderation_rule(cls, guild_id: int, rule_id: int) -> str:
        return f"guilds/{guild_id}/auto-moderation/rules/{rule_id}"

    @classmethod
    def create_auto_moderation_rule(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/auto-moderation/rules"

    @classmethod
    def modify_auto_moderation_rule(cls, guild_id: int, rule_id: int) -> str:
        return f"guilds/{guild_id}/auto-moderation/rules/{rule_id}"

    @classmethod
    def delete_auto_moderation_rule(cls, guild_id: int, rule_id: int) -> str:
        return f"guilds/{guild_id}/auto-moderation/rules/{rule_id}"

    @classmethod
    def get_channel(cls, channel_id: int) -> str:
        return f"channels/{channel_id}"

    @classmethod
    def modify_channel(cls, channel_id: int) -> str:
        return f"channels/{channel_id}"

    @classmethod
    def delete_channel(cls, channel_id: int) -> str:
        return f"channels/{channel_id}"

    @classmethod
    def get_channel_messages(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/messages"

    @classmethod
    def get_channel_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}"

    @classmethod
    def create_message(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/messages"

    @classmethod
    def crosspost_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}/crosspost"

    @classmethod
    def create_reaction(cls, channel_id: int, message_id: int, emoji: str) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me"

    @classmethod
    def delete_own_reaction(cls, channel_id: int, message_id: int, emoji: str) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me"

    @classmethod
    def delete_user_reaction(cls, channel_id: int, message_id: int, emoji: str, user_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{user_id}"

    @classmethod
    def get_user_reactions(cls, channel_id: int, message_id: int, emoji: str) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions/{emoji}"

    @classmethod
    def delete_all_reactions(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions"

    @classmethod
    def delete_all_reactions_for_emoji(cls, channel_id: int, message_id: int, emoji: str) -> str:
        return f"channels/{channel_id}/messages/{message_id}/reactions/{emoji}"

    @classmethod
    def edit_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}"

    @classmethod
    def delete_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}"

    @classmethod
    def bulk_delete_messages(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/messages/bulk-delete"

    @classmethod
    def edit_channel_permissions(cls, channel_id: int, overwrite_id: int) -> str:
        return f"channels/{channel_id}/permissions/{overwrite_id}"

    @classmethod
    def get_channel_invites(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/invites"

    @classmethod
    def create_channel_invite(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/invites"

    @classmethod
    def delete_channel_permission(cls, channel_id: int, overwrite_id: int) -> str:
        return f"channels/{channel_id}/permissions/{overwrite_id}"

    @classmethod
    def follow_news_channel(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/followers"

    @classmethod
    def trigger_typing_indicator(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/typing"

    @classmethod
    def get_pinned_messages(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/pins"

    @classmethod
    def add_pinned_channel_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/pins/{message_id}"

    @classmethod
    def delete_pinned_channel_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/pins/{message_id}"

    @classmethod
    def group_dm_add_recipient(cls, channel_id: int, user_id: int) -> str:
        return f"channels/{channel_id}/recipients/{user_id}"

    @classmethod
    def group_dm_remove_recipient(cls, channel_id: int, user_id: int) -> str:
        return f"channels/{channel_id}/recipients/{user_id}"

    @classmethod
    def start_thread_with_message(cls, channel_id: int, message_id: int) -> str:
        return f"channels/{channel_id}/messages/{message_id}/threads"

    @classmethod
    def start_thread_without_message(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/threads"

    @classmethod
    def start_thread_in_forum(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/threads"

    @classmethod
    def join_thread(cls, channel_id: int, thread_id: int) -> str:
        return f"channels/{channel_id}/threads/{thread_id}/members/@me"

    @classmethod
    def add_thread_member(cls, channel_id: int, thread_id: int, user_id: int) -> str:
        return f"channels/{channel_id}/threads/{thread_id}/members/{user_id}"

    @classmethod
    def leave_thread(cls, channel_id: int, thread_id: int) -> str:
        return f"channels/{channel_id}/threads/{thread_id}/members/@me"

    @classmethod
    def remove_thread_member(cls, channel_id: int, thread_id: int, user_id: int) -> str:
        return f"channels/{channel_id}/threads/{thread_id}/members/{user_id}"

    @classmethod
    def get_thread_member(cls, channel_id: int, thread_id: int, user_id: int) -> str:
        return f"channels/{channel_id}/threads/{thread_id}/members/{user_id}"

    @classmethod
    def list_thread_members(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/threads-members"

    @classmethod
    def list_public_archived_threads(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/threads/archived/public"

    @classmethod
    def list_private_archived_threads(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/threads/archived/private"

    @classmethod
    def list_joined_private_archived_threads(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/users/@me/threads/archived/private"

    @classmethod
    def list_guild_emojis(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/emojis"

    @classmethod
    def get_guild_emoji(cls, guild_id: int, emoji_id: int) -> str:
        return f"guilds/{guild_id}/emojis/{emoji_id}"

    @classmethod
    def create_guild_emoji(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/emojis"

    @classmethod
    def modify_guild_emoji(cls, guild_id: int, emoji_id: int) -> str:
        return f"guilds/{guild_id}/emojis/{emoji_id}"

    @classmethod
    def delete_guild_emoji(cls, guild_id: int, emoji_id: int) -> str:
        return f"guilds/{guild_id}/emojis/{emoji_id}"

    @classmethod
    def create_guild(cls) -> str:
        return "guilds"

    @classmethod
    def get_guild(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}"

    @classmethod
    def get_guild_preview(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/preview"

    @classmethod
    def edit_guild(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}"

    @classmethod
    def delete_guild(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}"

    @classmethod
    def get_guild_channels(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/channels"

    @classmethod
    def create_guild_channel(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/channels"

    @classmethod
    def modify_guild_channel_positions(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/channels"

    @classmethod
    def active_guild_threads(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/threads/active"

    @classmethod
    def get_guild_member(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}"

    @classmethod
    def list_guild_members(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/members"

    @classmethod
    def search_guild_members(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/members/search"

    @classmethod
    def add_guild_member(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}"

    @classmethod
    def modify_guild_member(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}"

    @classmethod
    def modify_current_member(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/members/@me"

    @classmethod
    def guild_member_addrole(cls, guild_id: int, user_id: int, role_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}/roles/{role_id}"

    @classmethod
    def guild_member_removerole(cls, guild_id: int, user_id: int, role_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}/roles/{role_id}"

    @classmethod
    def remove_guild_member(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/members/{user_id}"

    @classmethod
    def get_guild_bans(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/bans"

    @classmethod
    def get_guild_ban(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/bans/{user_id}"

    @classmethod
    def create_guild_ban(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/bans/{user_id}"

    @classmethod
    def remove_guild_ban(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/bans/{user_id}"

    @classmethod
    def get_guild_roles(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/roles"

    @classmethod
    def create_guild_role(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/roles"

    @classmethod
    def modify_guild_role_positions(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/roles"

    @classmethod
    def modify_guild_role(cls, guild_id: int, role_id: int) -> str:
        return f"guilds/{guild_id}/roles/{role_id}"

    @classmethod
    def modify_guild_mfa(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/mfa"

    @classmethod
    def delete_guild_role(cls, guild_id: int, role_id: int) -> str:
        return f"guilds/{guild_id}/roles/{role_id}"

    @classmethod
    def get_guild_prune_count(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/prune"

    @classmethod
    def begin_guild_prune(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/prune"

    @classmethod
    def guild_voice_regions(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/regions"

    @classmethod
    def get_guild_invites(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/invites"

    @classmethod
    def get_guild_integrations(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/integrations"

    @classmethod
    def delete_guild_integration(cls, guild_id: int, integration_id: int) -> str:
        return f"guilds/{guild_id}/integrations/{integration_id}"

    @classmethod
    def get_guild_widget_settings(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/widget"

    @classmethod
    def modify_guild_widget(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/widget"

    @classmethod
    def get_guild_widget(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/widget.json"

    @classmethod
    def guild_vanity_url(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/vanity-url"

    @classmethod
    def get_guild_widget_image(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/widget.png"

    @classmethod
    def get_guild_welcome_screen(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/welcome-screen"

    @classmethod
    def modify_guild_welcome_screen(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/welcome-screen"

    @classmethod
    def modify_current_user_voice_state(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/voice-states/@me"

    @classmethod
    def modify_user_voice_state(cls, guild_id: int, user_id: int) -> str:
        return f"guilds/{guild_id}/voice-states/{user_id}"

    @classmethod
    def list_scheduled_guild_events(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events"

    @classmethod
    def create_scheduled_guild_event(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events"

    @classmethod
    def get_scheduled_guild_event(cls, guild_id: int, event_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events/{event_id}"

    @classmethod
    def modify_scheduled_guild_event(cls, guild_id: int, event_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events/{event_id}"

    @classmethod
    def delete_scheduled_guild_event(cls, guild_id: int, event_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events/{event_id}"

    @classmethod
    def get_scheduled_guild_event_users(cls, guild_id: int, event_id: int) -> str:
        return f"guilds/{guild_id}/scheduled-events/{event_id}/users"

    @classmethod
    def get_guild_template(cls, guild_id: int, template_code: str) -> str:
        return f"guilds/{guild_id}/templates/{template_code}"

    @classmethod
    def create_guild_from_template(cls, guild_id: int, template_code: str) -> str:
        return f"guilds/{guild_id}/templates/{template_code}"

    @classmethod
    def get_guild_templates(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/templates"

    @classmethod
    def create_guild_template(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/templates"

    @classmethod
    def sync_guild_template(cls, guild_id: int, template_code: str) -> str:
        return f"guilds/{guild_id}/templates/{template_code}"

    @classmethod
    def modify_guild_template(cls, guild_id: int, template_code: str) -> str:
        return f"guilds/{guild_id}/templates/{template_code}"

    @classmethod
    def delete_guild_template(cls, guild_id: int, template_code: str) -> str:
        return f"guilds/{guild_id}/templates/{template_code}"

    @classmethod
    def get_invite(cls, invite_code: str) -> str:
        return f"invites/{invite_code}"

    @classmethod
    def delete_invite(cls, invite_code: str) -> str:
        return f"invites/{invite_code}"

    @classmethod
    def create_stage_instance(cls) -> str:
        return "stage-instances"

    @classmethod
    def get_stage_instance(cls, channel_id: int) -> str:
        return f"stage-instances/{channel_id}"

    @classmethod
    def modify_stage_instance(cls, channel_id: int) -> str:
        return f"stage-instances/{channel_id}"

    @classmethod
    def delete_stage_instance(cls, channel_id: int) -> str:
        return f"stage-instances/{channel_id}"

    @classmethod
    def get_sticker(cls, sticker_id: int) -> str:
        return f"stickers/{sticker_id}"

    @classmethod
    def list_nitro_sticker_packs(cls) -> str:
        return "sticker-packs"

    @classmethod
    def list_guild_stickers(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/stickers"

    @classmethod
    def create_guild_sticker(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/stickers"

    @classmethod
    def get_guild_sticker(cls, guild_id: int, sticker_id: int) -> str:
        return f"guilds/{guild_id}/stickers/{sticker_id}"

    @classmethod
    def modify_guild_sticker(cls, guild_id: int, sticker_id: int) -> str:
        return f"guilds/{guild_id}/stickers/{sticker_id}"

    @classmethod
    def delete_guild_sticker(cls, guild_id: int, sticker_id: int) -> str:
        return f"guilds/{guild_id}/stickers/{sticker_id}"

    @classmethod
    def get_current_user(cls) -> str:
        return "users/@me"

    @classmethod
    def get_user(cls, user_id: int) -> str:
        return f"users/{user_id}"

    @classmethod
    def modify_current_user(cls) -> str:
        return "users/@me"

    @classmethod
    def get_current_user_guilds(cls) -> str:
        return "users/@me/guilds"

    @classmethod
    def get_current_user_guilds_membership(cls, guild_id: int) -> str:
        return f"users/@me/guilds/{guild_id}/member"

    @classmethod
    def leave_guild(cls, guild_id: int) -> str:
        return f"users/@me/guilds/{guild_id}"

    @classmethod
    def create_dm(cls) -> str:
        return "users/@me/channels"

    @classmethod
    def create_group_dm(cls) -> str:
        return "users/@me/channels"

    @classmethod
    def get_user_connections(cls) -> str:
        return "users/@me/connections"

    @classmethod
    def list_voice_regions(cls) -> str:
        return "voice/regions"

    @classmethod
    def create_webhook(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/webhooks"

    @classmethod
    def get_channel_webhooks(cls, channel_id: int) -> str:
        return f"channels/{channel_id}/webhooks"

    @classmethod
    def get_webhook(cls, webhook_id: int) -> str:
        return f"webhooks/{webhook_id}"

    @classmethod
    def get_webhook_with_token(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}"

    @classmethod
    def modify_webhook(cls, webhook_id: int) -> str:
        return f"webhooks/{webhook_id}"

    @classmethod
    def modify_webhook_with_token(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}"

    @classmethod
    def delete_webhook(cls, webhook_id: int) -> str:
        return f"webhooks/{webhook_id}"

    @classmethod
    def delete_webhook_with_token(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}"

    @classmethod
    def execute_webhook(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}"

    @classmethod
    def execute_slack_compatible_webhook(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}/slack"

    @classmethod
    def execute_github_compatible_webhook(cls, webhook_id: int, webhook_token: str) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}/github"

    @classmethod
    def get_webhook_message(cls, webhook_id: int, webhook_token: str, message_id: int) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"

    @classmethod
    def edit_webhook_message(cls, webhook_id: int, webhook_token: str, message_id: int) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"

    @classmethod
    def delete_webhook_message(cls, webhook_id: int, webhook_token: str, message_id: int) -> str:
        return f"webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"

    @classmethod
    def get_guild_webhooks(cls, guild_id: int) -> str:
        return f"guilds/{guild_id}/webhooks"

    @classmethod
    def interaction_command(cls, app_id: int) -> str:
        return f"applications/{app_id}/commands"

    @classmethod
    def interaction_callback(cls, interaction_id: int, interaction_token: str) -> str:
        return f"interactions/{interaction_id}/{interaction_token}/callback"

    @classmethod
    def get_original_interaction(cls, interaction_id: int, interaction_token: str) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/@original"

    @classmethod
    def edit_original_interaction(cls, interaction_id: int, interaction_token: str) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/@original"

    @classmethod
    def delete_original_interaction(cls, interaction_id: int, interaction_token: str) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/@original"

    @classmethod
    def create_followup_message(cls, interaction_id: int, interaction_token: str) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}"

    @classmethod
    def get_followup_message(cls, interaction_id: int, interaction_token: str, message_id: int) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/{message_id}"

    @classmethod
    def edit_followup_message(cls, interaction_id: int, interaction_token: str, message_id: int) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/{message_id}"

    @classmethod
    def delete_followup_message(cls, interaction_id: int, interaction_token: str, message_id: int) -> str:
        return f"/webhooks/{interaction_id}/{interaction_token}/messages/{message_id}"



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import enum
import typing

import attrs

from wyvern._internals import _ListenerArg

if typing.TYPE_CHECKING:
    from wyvern import models
    from wyvern.clients import GatewayClient
    from wyvern.components import ActionRowContainer, Modal
    from wyvern.constructors.embeds import EmbedConstructor


__all__: tuple[str, ...] = (
    "Interaction",
    "InteractionCommandOptionType",
    "InteractionCommandType",
    "InteractionType",
    "InteractionResponseType",
)


@typing.final
class InteractionType(enum.IntEnum):
    """Type of the interaction."""

    PING = 1
    """A ping-pong interaction."""
    APPLICATION_COMMAND = 2
    """Interaction recieved on application command invocation"""
    MESSAGE_COMPONENT = 3
    """Interaction recieved when a message component is used."""
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    """Interaction for autocompletes."""
    MODAL_SUBMIT = 5
    """Interaction when a modal form is submitted."""


@typing.final
class InteractionResponseType(enum.IntEnum):
    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    MODAL = 9


@typing.final
class InteractionCommandType(enum.IntEnum):
    CHAT_INPUT = 1
    USER = 2
    MESSAGE = 3


@typing.final
class InteractionCommandOptionType(enum.IntEnum):
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10
    ATTACHMENT = 11


@attrs.define(kw_only=True, slots=True, repr=True, init=True)
class Interaction(_ListenerArg):
    """Represents a discord interaction."""

    client: "GatewayClient"
    id: int
    """ID of the interaction."""
    application_id: models.base.Snowflake
    """ID of the application that recieved the interaction."""
    type: InteractionType
    """Type of the interaction."""
    data: typing.Any
    """Interaction data."""
    guild_id: int | None
    """ID of the guild where interaction was created."""
    channel_id: int | None
    """ID of channel where interaction was created."""
    user: models.users.User | None
    """User who invoked the interaction."""
    token: str
    """The interaction token."""
    version: int = 1
    message: "models.messages.Message" | None
    """The message related to this interaction."""
    guild_locale: str | None
    """Locale of the guild."""

    async def create_message_response(
        self,
        content: str | None = None,
        *,
        embed: "EmbedConstructor" | None = None,
        embeds: typing.Sequence["EmbedConstructor"] = (),
        components: typing.Sequence["ActionRowContainer"] = (),
        allowed_mentions: "models.AllowedMentions" | None = None,
    ) -> None:
        if all([embed, embeds]):
            raise ValueError("You cannot use both embed and embeds kwarg.")
        embeds = [embed] if embed else embeds
        await self.client.rest.create_interaction_response(
            self,
            InteractionResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
            content=content,
            embeds=embeds,
            components=components,
            allowed_mentions=allowed_mentions,
        )

    async def create_defered_response(self) -> None:
        await self.client.rest.create_interaction_response(
            self,
            InteractionResponseType.DEFERRED_UPDATE_MESSAGE
            if self.type is InteractionType.MESSAGE_COMPONENT
            else InteractionResponseType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE,
        )

    async def create_modal_response(self, modal: "Modal") -> None:
        await self.client.rest.create_interaction_response(self, InteractionResponseType.MODAL, modal=modal)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import typing

from wyvern.files import Attachment
from wyvern.models import Snowflake, _converters

from .applications import (
    ApplicationCommandInteraction,
    ApplicationCommandInteractionData,
    ApplicationCommandInteractionResolvedData,
    InteractionOption,
)
from .base import Interaction, InteractionType
from .components import ComponentInteraction, ComponentInteractionData

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient


def payload_to_option(data: dict[str, typing.Any]) -> InteractionOption:
    return InteractionOption(
        name=data["name"],
        type=data["type"],
        value=data.get("value"),
        options=[payload_to_option(_data) for _data in data.get("options", [])],
        is_focused=data.get("focused", False),
    )


def payload_to_resolved(
    client: "GatewayClient", guild_id: int | None, data: dict[str, typing.Any]
) -> ApplicationCommandInteractionResolvedData:
    users = [_converters.payload_to_user(client, payload=_data) for _data in data.get("users", {}).values()]
    members = []
    messages = [_converters.payload_to_message(client, _data) for _data in data.get("messages", {}).values()]
    attachments = [Attachment._from_payload(client, Snowflake, _data) for _data in data.get("attachments", {}).values()]
    channels = [_converters.payload_to_channel(client, _data) for _data in data.get("channels", {}).values()]
    if guild_id is not None:
        members = [
            _converters.payload_to_member(client, guild_id=Snowflake.create(guild_id), payload=_data)
            for _data in data.get("members", {}).values()
        ]

    res = ApplicationCommandInteractionResolvedData(
        users={u.id: u for u in users},
        members={m.id: m for m in members},
        messages={m.id: m for m in messages},
        attachments={a.id: a for a in attachments},
        channels={c.id: c for c in channels},
    )
    return res


def payload_to_interaction(client: "GatewayClient", payload: dict[str, typing.Any]) -> Interaction:
    inter: Interaction
    if payload["type"] == InteractionType.MESSAGE_COMPONENT:
        print(payload["data"])
        inter = ComponentInteraction(
            client=client,
            id=payload["id"],
            application_id=payload["application_id"],
            type=payload["type"],
            data=ComponentInteractionData(**payload["data"]),
            token=payload["token"],
            version=payload["version"],
            message=_converters.payload_to_message(client, msg) if (msg := payload.get("message")) else None,
            user=_converters.payload_to_user(client, user) if (user := payload.get("user")) else None,
            guild_id=payload.get("guild_id"),
            channel_id=payload.get("guild_id"),
            guild_locale=payload.get("guild_locale"),
        )
    elif payload["type"] == InteractionType.APPLICATION_COMMAND:
        inter = ApplicationCommandInteraction(
            client=client,
            id=payload["id"],
            application_id=payload["application_id"],
            token=payload["token"],
            version=payload["version"],
            message=_converters.payload_to_message(client, msg) if (msg := payload.get("message")) else None,
            type=InteractionType.APPLICATION_COMMAND,
            data=ApplicationCommandInteractionData(
                payload=payload["data"],
                command_id=((d := payload["data"])["id"]),
                command_name=d["name"],
                guild_id=d.get("guild_id"),
                command_type=d["type"],
                target_id=d.get("target_id"),
                options=[payload_to_option(data) for data in payload.get("options", [])],
                resoloved=payload_to_resolved(client, payload.get("guild_id"), data)
                if (data := payload.get("resolved"))
                else None,
            ),
            guild_id=payload.get("guild_id"),
            channel_id=payload.get("channel_id"),
            user=_converters.payload_to_user(client, user) if (user := payload.get("user")) else None,
            guild_locale=payload.get("guild_locale"),
        )

    else:
        pass

    return inter  # type: ignore



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import colorsys
import random
import re
import typing

__all__: tuple[str, ...] = ("Color",)


@typing.final
class Color:
    """
    Class representing a color in the RGB color space.
    Alias name Colour exists for convenience.

    Attributes
    ----------
    value : int
        The value of the color. This is a 24-bit integer, where the first 8 bits are the red value,
        the next 8 bits are the green value, and the last 8 bits are the blue value.
    """

    __slots__: tuple[str, ...] = ("value",)

    RGB_REGEX: re.Pattern[str] = re.compile(r"rgb\((\d{1,3}), (\d{1,3}), (\d{1,3})\)")
    HSL_REGEX: re.Pattern[str] = re.compile(r"hsl\((\d{1,3}), (\d{1,3})%, (\d{1,3})%\)")
    HSV_REGEX: re.Pattern[str] = re.compile(r"hsv\((\d{1,3}), (\d{1,3})%, (\d{1,3})%\)")
    HEX_REGEX: re.Pattern[str] = re.compile(r"#([0-9a-fA-F]{3}|[0-9a-fA-F]{6})")

    def __init__(self, value: int) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f"Color({self.value})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Color) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __ne__(self, other: object) -> bool:
        return self is not other

    @classmethod
    def from_hex(cls, hex_value: str) -> Color:
        """
        Creates a Color object from a hex value.

        Parameters
        ----------

        hex_value: str
            The hex value to use.

        Returns
        -------

        wyvern.Color
            A Color object.

        Examples
        --------

            >>> Color.from_hex('#ff0000')
            Color(16776960)
            >>> Color.from_hex('#00ff00')
            Color(255)
            >>> Color.from_hex('#0000ff')
            Color(0)
        """
        if match := cls.HEX_REGEX.match(hex_value):
            hex_value = match.group(1)
            if len(hex_value) == 3:
                hex_value = "".join(c * 2 for c in hex_value)
            return cls(int(hex_value, 16))
        raise ValueError(f"Invalid hex value: {hex_value}")

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int) -> Color:
        """
        Creates a Color object from RGB values.

        Parameters
        ----------
        r: int
            The red value.
        g: int
            The green value.
        b: int
            The blue value.

        Returns
        -------
        wyvern.Color
            A Color object.

        Examples
        --------

            >>> Color.from_rgb(255, 0, 0)
            Color(16711680)
            >>> Color.from_rgb(0, 255, 0)
            Color(65280)
            >>> Color.from_rgb(0, 0, 255)
            Color(255)
        """
        return cls((r << 16) + (g << 8) + b)

    @classmethod
    def from_hsv(cls, h: float, s: float, v: float) -> Color:
        """
        Creates a Color object from HSV values.

        Parameters
        ----------
        h: float
            The hue value.
        s: float
            The saturation value.
        v: float
            The value in HSV color space.

        Returns
        -------
        wyvern.Color
            A Color object.

        Examples
        --------

            >>> Color.from_hsv(0, 1, 1)
            Color(16711680)
            >>> Color.from_hsv(120, 1, 1)
            Color(16711680)
            >>> Color.from_hsv(240, 1, 1)
            Color(16711680)

        """
        return cls.from_rgb(*[int(round(c * 255)) for c in colorsys.hsv_to_rgb(h, s, v)])

    @classmethod
    def from_hsl(cls, h: float, s: float, l: float) -> Color:  # noqa: E741
        """
        Creates a Color object from HSL values.

        Parameters
        ----------
        h: float
            The hue value.
        s: float
            The saturation value.
        l: float
            The lightness value.

        Returns
        -------
        wyvern.Color
            A Color object.

        Examples
        --------

            >>> Color.from_hsl(0, 1, 0.5)
            Color(16711680)
            >>> Color.from_hsl(120, 1, 0.5)
            Color(16711680)
            >>> Color.from_hsl(240, 1, 0.5)
            Color(16711680)
        """
        return cls.from_rgb(*[int(round(c * 255)) for c in colorsys.hls_to_rgb(h, l, s)])

    @classmethod
    def from_random(cls) -> Color:
        """
        Creates a Color object from a random color. Randomly generates a color in the RGB color space.

        Returns
        -------

        wyvern.Color
            A Color object.
        """
        return cls.from_rgb(*[random.randint(0, 255) for _ in range(3)])

    @classmethod
    def from_string(cls, string: str) -> Color:
        """
        Creates a Color object from a string.

        Parameters
        ----------

        string: str
            The string to use.

        Returns
        -------

        wyvern.Color
            A Color object.

        Examples
        --------

            >>> Color.from_string('rgb(255, 0, 0)')
            Color(16711680)
            >>> Color.from_string('hsl(0, 100%, 50%)')
            Color(-80727249750)
            >>> Color.from_string('hsv(0, 100%, 100%)')
            Color(1022371500)
            >>> Color.from_string('#ff0000')
            Color(16776960)

        """
        if string.startswith("#"):
            return cls.from_hex(string)
        elif match := cls.RGB_REGEX.match(string):
            return cls.from_rgb(*[int(c) for c in match.groups()])
        elif match := cls.HSL_REGEX.match(string):
            return cls.from_hsl(*[float(c) for c in match.groups()])
        elif match := cls.HSV_REGEX.match(string):
            return cls.from_hsv(*[float(c) for c in match.groups()])
        raise ValueError(f"Invalid color string: {string}")

    @classmethod
    def default(cls) -> Color:
        """
        Creates a Color object from the default color. This is `0x000000`. (Black)

        Examples
        --------

            >>> Color.default()
            Color(0)

        """
        return cls(0x000000)

    @property
    def hex(self) -> str:
        """The hex value of the color."""
        return f"#{self.value:06x}"

    @property
    def rgb(self) -> tuple[int, int, int]:
        """The RGB values of the color."""
        return (self.value >> 16) & 0xFF, (self.value >> 8) & 0xFF, self.value & 0xFF

    @property
    def hsv(self) -> tuple[float, float, float]:
        """The HSV values of the color."""
        return colorsys.rgb_to_hsv(*(c / 255 for c in self.rgb))

    @property
    def hsl(self) -> tuple[float, float, float]:
        """The HSL values of the color."""
        return colorsys.rgb_to_hls(*(c / 255 for c in self.rgb))

    @property
    def r(self) -> int:
        """The red value of the color."""
        return self.rgb[0]

    @property
    def g(self) -> int:
        """The green value of the color."""
        return self.rgb[1]

    @property
    def b(self) -> int:
        """The blue value of the color."""
        return self.rgb[2]

    @classmethod
    def red(cls) -> Color:
        """Creates a Color object from the red color. This is `0xff0000`. (Red)"""
        return cls(0xFF0000)

    @classmethod
    def green(cls) -> Color:
        """Creates a Color object from the green color. This is `0x00ff00`. (Green)"""
        return cls(0x00FF00)

    @classmethod
    def blue(cls) -> Color:
        """Creates a Color object from the blue color. This is `0x0000ff`. (Blue)"""
        return cls(0x0000FF)

    @classmethod
    def yellow(cls) -> Color:
        """Creates a Color object from the yellow color. This is `0xffff00`. (Yellow)"""
        return cls(0xFFFF00)

    @classmethod
    def cyan(cls) -> Color:
        """Creates a Color object from the cyan color. This is `0x00ffff`. (Cyan)"""
        return cls(0x00FFFF)

    @classmethod
    def magenta(cls) -> Color:
        """Creates a Color object from the magenta color. This is `0xff00ff`. (Magenta)"""
        return cls(0xFF00FF)

    @classmethod
    def black(cls) -> Color:
        """Creates a Color object from the black color. This is `0x000000`. (Black)"""
        return cls(0x000000)

    @classmethod
    def white(cls) -> Color:
        """Creates a Color object from the white color. This is `0xffffff`. (White)"""
        return cls(0xFFFFFF)

    @classmethod
    def gray(cls) -> Color:
        """Creates a Color object from the gray color. This is `0x808080`. (Gray)"""
        return cls(0x808080)

    @classmethod
    def grey(cls) -> Color:
        """Creates a Color object from the grey color. This is `0x808080`. (Grey)"""
        return cls(0x808080)

    @classmethod
    def orange(cls) -> Color:
        """Creates a Color object from the orange color. This is `0xffa500`. (Orange)"""
        return cls(0xFFA500)

    @classmethod
    def purple(cls) -> Color:
        """Creates a Color object from the purple color. This is `#800080`. (Purple)"""
        return cls(0x800080)

    @classmethod
    def brown(cls) -> Color:
        """Creates a Color object from the brown color. This is `#a52a2a`. (Brown)"""
        return cls(0xA52A2A)

    @classmethod
    def silver(cls) -> Color:
        """Creates a Color object from the silver color. This is `#c0c0c0`. (Silver)"""
        return cls(0xC0C0C0)

    @classmethod
    def aqua(cls) -> Color:
        """Creates a Color object from the aqua color. This is `#00ffff`. (Aqua)"""
        return cls(0x00FFFF)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.models.users import User

__all__: tuple[str, ...] = ("UsersState",)


class UsersState:
    """A handler for user cache and rest connection with the client.
    This interface can be used to get/fetch users with convinience.
    """

    _client: "GatewayClient"
    cached_users: dict[int, "User"] = {}
    """Mapping of cached users with their ID : Object"""

    def __init__(self, client: "GatewayClient") -> None:
        self._client = client

    def get(self, user_id: int) -> "User" | None:
        """Fetchs a user using the state cache..

        Parameters
        ----------

        user_id : int
            ID of the user that is to be to get.

        Returns
        -------

        wyvern.models.users.User | None
            The user object that was fetched.

        """
        return self.cached_users.get(user_id)

    async def fetch(self, user_id: int) -> "User":
        """Fetchs a user using the REST api.

        Parameters
        ----------

        user_id : int
            ID of the user that is to be fetched.

        Returns
        -------

        wyvern.models.users.User
            The user object that was fetched.

        Raises
        ------

        wyvern.exceptions.UserNotFound
            The targetted user was not found.
        """
        return await self._client.rest.fetch_user(user_id)

    def get_user_named(self, name: str) -> "User" | None:
        """Gets a member named the provided argument from cache

        Parameters
        ----------

        name : str
            Name to lookup for.

        Returns
        -------

        wyvern.models.users.User | None
            The user object that was fetched.

        """
        users = [user for user in self.cached_users.values() if user.username == name]
        return users[0] if users != [] else None

    def parse_from_string(self, string: str) -> "User" | None:
        """
        Parses a user object from a string

        The order for parsing:

        * Lookup for mentions.
        * Lookup for username#discriminator.
        * Lookup for username only.

        Parameters
        ----------

        string : str
            The string to parse from.

        Returns
        -------

        wyvern.models.users.User | None
            The user object that was parsed, if any.

        """
        if string.startswith("<@") and string.endswith(">"):
            return self.get(int(string.strip("<@!>")))
        elif "#" in string:
            username, discriminator = string.split("#")
            users = [
                user
                for user in self.cached_users.values()
                if user.username == username and int(user.discriminator) == int(discriminator)
            ]
            if users != []:
                return users[0]
            else:
                pass
        elif user := self.get_user_named(string):
            return user
        return None

    def _add_to_state(self, *users: "User") -> None:
        for user in users:
            self.cached_users[user.id] = user



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

from wyvern._internals import BitWiseFlag

__all__: tuple[str, ...] = ("Intents",)


@typing.final
class Intents(BitWiseFlag):
    """
    Intents constructor to provide to the gateway.

    Attributes
    ----------

    value: int
        The final value of intents generated by the constructor.

    """

    NONE = 0
    """
    No intents.
    """
    GUILDS = 1 << 0
    """Required for these gateway events:

    * `GUILD_CREATE`
    * `GUILD_UPDATE`
    * `GUILD_DELETE`
    * `GUILD_ROLE_CREATE`
    * `GUILD_ROLE_UPDATE`
    * `GUILD_ROLE_DELETE`
    * `CHANNEL_CREATE`
    * `CHANNEL_UPDATE`
    * `CHANNEL_DELETE`
    * `CHANNEL_PINS_UPDATE`
    * `THREAD_CREATE`
    * `THREAD_UPDATE`
    * `THREAD_DELETE`
    * `THREAD_LIST_SYNC`
    * `THREAD_MEMBER_UPDATE`
    * `THREAD_MEMBERS_UPDATE`
    * `STAGE_INSTANCE_CREATE`
    * `STAGE_INSTANCE_UPDATE`
    * `STAGE_INSTANCE_DELETE`
    """
    GUILD_MEMBERS = 1 << 1
    """Required for these gateway events:

    * `GUILD_MEMBER_ADD`
    * `GUILD_MEMBER_UPDATE`
    * `GUILD_MEMBER_REMOVE`
    * `THREAD_MEMBERS_UPDATE`

    !!! warning
        This is a privileged intent.
    """
    GUILD_BANS = 1 << 2
    """Required for these gateway events:

    * `GUILD_BAN_ADD`
    * `GUILD_BAN_REMOVE`
    """
    GUILD_EMOJIS = 1 << 3
    """Required for these gateway events:

    * `GUILD_EMOJIS_UPDATE`
    * `GUILD_STICKERS_UPDATE`
    """
    GUILD_INTEGRATIONS = 1 << 4
    """Required for these gateway events:

    * `GUILD_INTEGRATIONS_UPDATE`
    * `INTEGRATION_CREATE`
    * `INTEGRATION_UPDATE`
    * `INTEGRATION_DELETE`
    """
    GUILD_WEBHOOKS = 1 << 5
    """Required for these gateway events:

    * `WEBHOOKS_UPDATE`
    """
    GUILD_INVITES = 1 << 6
    """Required for these gateway events:

    * `INVITE_CREATE`
    * `INVITE_DELETE`
    """
    GUILD_VOICE_STATES = 1 << 7
    """Required for these gateway events:

    * `VOICE_STATE_UPDATE`
    """
    GUILD_PRESENCES = 1 << 8
    """Required for these gateway events:

    * `PRESENCE_UPDATE`

    !!! warning
        This is a privileged intent.
    """
    GUILD_MESSAGES = 1 << 9
    """Required for these gateway events:

    * `MESSAGE_CREATE`
    * `MESSAGE_UPDATE`
    * `MESSAGE_DELETE`
    * `MESSAGE_DELETE_BULK`
    """
    GUILD_MESSAGE_REACTIONS = 1 << 10
    """Required for these gateway events:

    * `MESSAGE_REACTION_ADD`
    * `MESSAGE_REACTION_REMOVE`
    * `MESSAGE_REACTION_REMOVE_ALL`
    * `MESSAGE_REACTION_REMOVE_EMOJI`
    """
    GUILD_MESSAGE_TYPING = 1 << 11
    """Required for these gateway events:

    * `TYPING_START`
    """
    DIRECT_MESSAGES = 1 << 12
    """Required for these gateway events:

    * `MESSAGE_CREATE`
    * `MESSAGE_UPDATE`
    * `MESSAGE_DELETE`
    * `CHANNEL_PINS_UPDATE`
    """
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    """Required for these gateway events:

    * `MESSAGE_REACTION_ADD`
    * `MESSAGE_REACTION_REMOVE`
    * `MESSAGE_REACTION_REMOVE_ALL`
    * `MESSAGE_REACTION_REMOVE_EMOJI`
    """
    DIRECT_MESSAGE_TYPING = 1 << 14
    """Required for these gateway events:

    * `TYPING_START`
    """
    MESSAGE_CONTENT = 1 << 15
    """Required for guild message's content.

    !!! warning
        This is a privileged intent.
    """
    GUILD_SCHEDULED_EVENTS = 1 << 16
    """Required for these gateway events:

    * `GUILD_SCHEDULED_EVENT_CREATE`
    * `GUILD_SCHEDULED_EVENT_UPDATE`
    * `GUILD_SCHEDULED_EVENT_DELETE`
    * `GUILD_SCHEDULED_EVENT_USER_ADD`
    * `GUILD_SCHEDULED_EVENT_USER_REMOVE`
    """
    AUTO_MODERATION_CONFIGURATION = 1 << 20
    """Required for these gateway events:

    * `AUTO_MODERATION_RULE_CREATE`
    * `AUTO_MODERATION_RULE_UPDATE`
    * `AUTO_MODERATION_RULE_DELETE`
    """

    AUTO_MODERATION_EXECUTION = 1 << 21
    """Required for these gateway events:

    * `AUTO_MODERATION_ACTION_EXECUTION`
    """

    UNPRIVILEGED = (
        GUILDS
        | GUILD_EMOJIS
        | GUILD_INTEGRATIONS
        | GUILD_WEBHOOKS
        | GUILD_INVITES
        | GUILD_VOICE_STATES
        | GUILD_MESSAGE_REACTIONS
        | GUILD_MESSAGE_TYPING
        | GUILD_MESSAGES
        | DIRECT_MESSAGES
        | DIRECT_MESSAGE_TYPING
        | DIRECT_MESSAGE_REACTIONS
        | AUTO_MODERATION_CONFIGURATION
        | AUTO_MODERATION_EXECUTION
    )
    """All unprivileged intents."""
    PRIVILEGED = MESSAGE_CONTENT | GUILD_MEMBERS | GUILD_PRESENCES
    """All privileged intents."""
    ALL = PRIVILEGED | UNPRIVILEGED
    """All intents."""



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import typing

__all__: tuple[str, ...] = ("BitWiseFlag", "_ListenerArg")


class BitWiseFlag:
    value: int

    def __init__(self, value: int) -> None:
        self.value = value

    _ignored_flags = ("ALL", "PRIVILEGED", "UNPRIVILEGED")

    def __iter__(self) -> typing.Generator[tuple[str, bool], None, None]:
        for flag in [attr for attr in dir(self) if attr.isupper() and attr not in self._ignored_flags]:
            if self.value & getattr(self, flag) != 0:
                yield flag, True
            else:
                yield flag, False

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(value={self.value})>"

    def __int__(self) -> int:
        return self.value

    def get_enabled(self, *, as_str: bool = False) -> tuple[int | str, ...]:
        return tuple((getattr(self, flag[0]) if as_str is False else flag[0]) for flag in list(self) if flag[1] is True)

    def get_disabled(self, *, as_str: bool = False) -> tuple[int | str, ...]:
        return tuple(
            (getattr(self, flag[0]) if as_str is False else flag[0]) for flag in list(self) if flag[1] is False
        )


class _ListenerArg:
    ...



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

from .base import ButtonStyle, Component, ComponentType
from .buttons import Button
from .selects import Select, SelectOption, SelectType

if typing.TYPE_CHECKING:
    from wyvern.models.channels import ChannelType
    from wyvern.models.emojis import CustomEmoji


__all__: tuple[str, ...] = ("ActionRowContainer",)


class ActionRowContainer(Component):
    """Interface to create an Action Row.
    This class accepts and stores. other discord components.
    """

    type = ComponentType.ACTION_ROW
    items: list[Component] = []
    """List of items stored in the container."""

    def __init__(self, *items: Component) -> None:
        for item in items:
            self.items.append(item)
        super().__init__()

    def to_payload(self) -> dict[str, typing.Any]:
        return {"type": int(self.type), "components": [item.to_payload() for item in self.items]}

    def add_select(
        self,
        *,
        type: SelectType = SelectType.STRING,
        custom_id: str,
        options: typing.Sequence[SelectOption] = (),
        channel_types: typing.Sequence[ChannelType] = (),
        min_values: int | None = None,
        max_values: int | None = None,
        disabled: bool = False,
    ) -> Select:
        """Adds a select menu to the container.

        Parameters
        ----------

        type: SelectType
            Type of the select
        custom_id: str
            Custom ID of the select.
        options: list[SelectOption]
            Options attached to the select.
        channel_types: list[ChannelType] | None
            Channel types for CHANNEL type select.
        min_values: int | None
            Minimum selections
        max_values: int | None
            Maximum selections
        disabled: bool
            True if the select is disabled.

        Returns
        -------

        Select
            The select that was created.
        """

        self.items.append(
            select := Select(
                type=type,
                custom_id=custom_id,
                options=list(options),
                channel_types=list(channel_types),
                min_values=min_values,
                max_values=max_values,
                disabled=disabled,
            )
        )
        return select

    def add_button(
        self,
        *,
        style: ButtonStyle = ButtonStyle.SECONDARY,
        label: str | None = None,
        emoji: str | CustomEmoji | None = None,
        custom_id: str | None = None,
        disabled: bool = False,
        url: str | None = None,
    ) -> Button:
        """Adds a button to the container.

        Parameters
        ----------

        style : wyvern.components.base.ButtonStyle
            The style of button.
        label : str | None
            Button's label.
        emoji : str | wyvern.models.emojis.CustomEmoji
            A unicode emoji or a custom one.
        custom_id : str
            A custom id set to the button.
        disabled : bool
            Set this to [True][] to disable clicks on button.
        url : str
            The URL this button points to.

        Returns
        -------

        wyvern.components.buttons.Button
            The button that was created.

        """
        self.items.append(
            button := Button(
                type=ComponentType.BUTTON,
                style=style,
                label=label,
                emoji=emoji,
                custom_id=custom_id,
                disabled=disabled,
                url=url,
            )
        )
        return button



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import typing

from wyvern._internals import _ListenerArg


def default_avatar_for(discriminator: int) -> str:
    return f"embed/avatars{discriminator % 5}.png"


__all__: tuple[str, ...] = ("DiscordObject", "Snowflake")


class Snowflake(int):
    @classmethod
    def create(cls, sf: typing.Any) -> "Snowflake":
        return cls(sf)


class DiscordObject(_ListenerArg):
    """
    Represents a discord object.

    Parameters
    ----------

    _id: int
        ID of the entity.

    Attributes
    ----------

    id: int
        ID of the entity.

    """

    def __init__(self, _id: int) -> None:
        self.id = Snowflake.create(_id)

    def __eq__(self, obj: object) -> bool:
        if not isinstance(obj, DiscordObject):
            return NotImplemented
        return self.id == obj.id

    @property
    def created_at(self) -> datetime.datetime:
        """
        The datetime at which this entity was created.

        Returns
        -------

        datetime.datetime
        """
        return self.get_created_at(self.id)

    @classmethod
    def get_created_at(cls, _id: int) -> datetime.datetime:
        """
        Get creation time of an entity using it's ID.
        """
        timestamp = ((_id >> 22) + 1420070400000) / 1000
        return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import enum
import typing

import attrs

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.models.base import Snowflake


class AvatarType(enum.Flag):
    """Represents avatar type."""

    GUILD = "GUILD"
    """Guild avatar."""
    CUSTOM = "CUSTOM"
    """Global custom avatar."""
    DEFAULT = "DEFAULT"
    """Default discord avatar."""


class BaseAsset:
    client: "GatewayClient"
    hash: str

    def __str__(self) -> str:
        return self.url

    @property
    def url(self) -> str:
        """Url of the asset."""
        return f"https://cdn.discordapp.com/{self.hash}.png"

    async def read(self) -> bytes:

        res = await self.client.rest._session.get(self.url)
        return await res.read()


@attrs.define(kw_only=True, slots=True, repr=True)
class Avatar(BaseAsset):
    """Represents an User's avatar."""

    client: "GatewayClient"
    type: AvatarType
    """Type of avatar."""
    hash: str
    """Hash value of avatar."""


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildIcon(BaseAsset):
    ...


@attrs.define(kw_only=True, slots=True, repr=True)
class Attachment(BaseAsset):
    """Represents a attachment in a message."""

    client: GatewayClient
    id: Snowflake
    """ID of the attachment."""
    filename: str
    """Attachments' filename."""
    description: str | None
    """Description, if any."""
    content_type: str | None
    """Type of the attachment."""
    size: int
    """Size of the attachment"""
    proxy_url: str
    """Proxy url of attachment"""
    height: int | None
    """Height of image."""
    width: int | None
    """Width of image."""
    is_ephemeral: bool | None
    """True if the attachement was ephemeral."""

    @property
    def is_spoiler(self) -> bool:
        """True if the attachment was spoiler enabled."""
        return True if self.filename.startswith("SPOILER_") else False

    @classmethod
    def _from_payload(
        cls: type["Attachment"], client: "GatewayClient", sfc: type["Snowflake"], data: dict[str, typing.Any]
    ) -> "Attachment":
        return cls(
            client=client,
            id=sfc.create(data["id"]),
            filename=data["filename"],
            description=data.get("description"),
            content_type=data.get("content_type"),
            size=data["size"],
            proxy_url=data["proxy_url"],
            height=data.get("height"),
            width=data.get("width"),
            is_ephemeral=data.get("ephemeral"),
        )



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import enum
import typing

import attrs

from wyvern._internals import BitWiseFlag

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from .channels import GuildChannel
    from .emojis import CustomEmoji
    from .roles import Role
    from .stickers import Sticker
    from .users import User


__all__: tuple[str, ...] = (
    "Guild",
    "GuildPreview",
    "GuildWidget",
    "GuildWidgetSettings",
    "WelcomeScreen",
    "WelcomeScreenChannel",
)


class GuildFeatures(enum.Enum):
    ANIMATED_ICON = "ANIMATED_ICON"
    BANNER = "BANNER"
    COMMERCE = "COMMERCE"
    COMMUNITY = "COMMUNITY"
    DISCOVERABLE = "DISCOVERABLE"
    FEATURABLE = "FEATURABLE"
    INVITE_SPLASH = "INVITE_SPLASH"
    MEMBER_VERIFICATION_GATE_ENABLED = "MEMBER_VERIFICATION_GATE_ENABLED"
    MONETIZATION_ENABLED = "MONETIZATION_ENABLED"
    MORE_EMOJI = "MORE_EMOJI"
    NEWS = "NEWS"
    PARTNERED = "PARTNERED"
    PREVIEW_ENABLED = "PREVIEW_ENABLED"
    PUBLIC = "PUBLIC"
    PUBLIC_DISABLED = "PUBLIC_DISABLED"
    RELAY_ENABLED = "RELAY_ENABLED"
    SEVEN_DAY_THREAD_ARCHIVE = "SEVEN_DAY_THREAD_ARCHIVE"
    THREE_DAY_THREAD_ARCHIVE = "THREE_DAY_THREAD_ARCHIVE"
    TICKETED_EVENTS_ENABLED = "TICKETED_EVENTS_ENABLED"
    VANITY_URL = "VANITY_URL"
    VERIFIED = "VERIFIED"
    VIP_REGIONS = "VIP_REGIONS"
    WELCOME_SCREEN_ENABLED = "WELCOME_SCREEN_ENABLED"


class MutableGuildFeatures(enum.Enum):
    COMMUNITY = "COMMUNITY"
    INVITES_DISABLED = "INVITES_DISABLED"
    DISCOVERABLE = "DISCOVERABLE"


class GuildVerificationLevel(enum.IntEnum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4


class GuildDefaultMessageNotificationLevel(enum.IntEnum):
    ALL_MESSAGES = 0
    ONLY_MENTIONS = 1


class GuildExplicitContentFilterLevel(enum.IntEnum):
    DISABLED = 0
    MEMBERS_WITHOUT_ROLES = 1
    ALL_MEMBERS = 2


class GuildMFALevel(enum.IntEnum):
    NONE = 0
    ELEVATED = 1


class GuildPremiumTier(enum.IntEnum):
    NONE = 0
    TIER_1 = 1
    TIER_2 = 2
    TIER_3 = 3


class GuildNSFWLevel(enum.IntEnum):
    DEFAULT = 0
    EXPLICIT = 1
    SAFE = 2
    AGE_RESTRICTED = 3


class GuildSystemChannelFlags(BitWiseFlag):
    SUPPRESS_JOIN_NOTIFICATIONS = 1 << 0
    SUPPRESS_PREMIUM_SUBSCRIPTIONS = 1 << 1
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS = 1 << 2
    SUPPRESS_JOIN_NOTIFICATION_REPLIES = 1 << 3


class IntegrationExpireBehavior(enum.IntEnum):
    REMOVE_ROLE = 0
    KICK = 1


@attrs.define(kw_only=True, slots=True, repr=True)
class Ban:
    reason: str | None
    user: User


@attrs.define(kw_only=True, slots=True, repr=True)
class IntegrationAccount:
    id: str
    name: str


@attrs.define(kw_only=True, slots=True, repr=True)
class IntegrationApplication(DiscordObject):
    name: str
    icon: typing.Optional[str]
    description: str
    summary: str
    bot: typing.Optional[User]


@attrs.define(kw_only=True, slots=True, repr=True)
class Integration(DiscordObject):
    raw: typing.Dict[str, typing.Any]
    id: Snowflake
    name: str
    type: str
    is_enabled: bool
    is_syncing: bool
    role_id: typing.Optional[Snowflake]
    enable_emoticons: typing.Optional[bool]
    expire_behavior: typing.Optional[IntegrationExpireBehavior]
    expire_grace_period: typing.Optional[int]
    user: typing.Optional[User]
    account: typing.Optional[IntegrationAccount]
    synced_at: typing.Optional[datetime.datetime]
    subscriber_count: typing.Optional[int]
    revoked: typing.Optional[bool]
    application: typing.Optional[IntegrationApplication]
    scopes: typing.Optional[typing.List[str]]

    @property
    def created_at(self) -> datetime.datetime:
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class WelcomeScreenChannel(DiscordObject):
    raw: dict[str, typing.Any]
    channel_id: Snowflake
    description: str
    emoji_id: typing.Optional[Snowflake]
    emoji_name: typing.Optional[str]


@attrs.define(kw_only=True, slots=True, repr=True)
class WelcomeScreen(DiscordObject):
    raw: dict[str, typing.Any]
    description: str
    welcome_channels: list[WelcomeScreenChannel]

    @property
    def channels(self) -> list[WelcomeScreenChannel]:
        return self.welcome_channels


@attrs.define(kw_only=True, slots=True, repr=True)
class Guild(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    name: str
    icon: typing.Optional[str]
    splash: typing.Optional[str]
    discovery_splash: typing.Optional[str]
    owner: bool
    owner_id: Snowflake
    permissions: str
    region: str
    afk_channel_id: typing.Optional[Snowflake]
    afk_timeout: int
    widget_enabled: bool
    widget_channel_id: typing.Optional[Snowflake]
    verification_level: GuildVerificationLevel
    default_message_notifications: GuildDefaultMessageNotificationLevel
    explicit_content_filter: GuildExplicitContentFilterLevel
    roles: list[Role]
    emojis: list[CustomEmoji]
    features: list[GuildFeatures]
    mfa_level: GuildMFALevel
    application_id: typing.Optional[Snowflake]
    system_channel_id: typing.Optional[Snowflake]
    system_channel_flags: GuildSystemChannelFlags
    rules_channel_id: typing.Optional[Snowflake]
    max_presences: typing.Optional[int]
    max_members: int
    vanity_url_code: typing.Optional[str]
    description: typing.Optional[str]
    banner: typing.Optional[str]
    premium_tier: GuildPremiumTier
    premium_subscription_count: int
    preferred_locale: str
    public_updates_channel_id: typing.Optional[Snowflake]
    max_video_channel_users: int
    approximate_member_count: int
    approximate_presence_count: int
    welcome_screen: typing.Optional[WelcomeScreen]
    nsfw_level: GuildNSFWLevel
    stickers: list[Sticker]
    premium_progress_bar_enabled: bool

    @property
    def icon_url(self) -> str | None:
        return f"https://cdn.discordapp.com/icons/{self.id}/{self.icon}.png" if self.icon else None

    @property
    def splash_url(self) -> str | None:
        return f"https://cdn.discordapp.com/splashes/{self.id}/{self.splash}.png" if self.splash else None

    @property
    def discovery_splash_url(self) -> str | None:
        return (
            f"https://cdn.discordapp.com/discovery-splashes/{self.id}/{self.discovery_splash}.png"
            if self.discovery_splash
            else None
        )

    @property
    def banner_url(self) -> str | None:
        return f"https://cdn.discordapp.com/banners/{self.id}/{self.banner}.png" if self.banner else None

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildPreview(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    name: str
    icon: typing.Optional[str]
    splash: typing.Optional[str]
    discovery_splash: typing.Optional[str]
    emojis: list[CustomEmoji]
    features: list[str]
    approximate_member_count: int
    approximate_presence_count: int
    description: typing.Optional[str]

    @property
    def icon_url(self) -> str | None:
        return f"https://cdn.discordapp.com/icons/{self.id}/{self.icon}.png" if self.icon else None

    @property
    def splash_url(self) -> str | None:
        return f"https://cdn.discordapp.com/splashes/{self.id}/{self.splash}.png" if self.splash else None

    @property
    def discovery_splash_url(self) -> str | None:
        return (
            f"https://cdn.discordapp.com/discovery-splashes/{self.id}/{self.discovery_splash}.png"
            if self.discovery_splash
            else None
        )

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildWidgetSettings(DiscordObject):
    raw: dict[str, typing.Any]
    enabled: bool
    channel_id: typing.Optional[Snowflake]


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildWidget(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    name: str
    instant_invite: typing.Optional[str]
    channels: list[GuildChannel]
    members: list[User]
    presence_count: int

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import abc
import datetime
import enum
import typing

import attrs

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.components.container import ActionRowContainer
    from wyvern.constructors.embeds import EmbedConstructor
    from wyvern.permissions import PermissionOverwrites

    from .messages import AllowedMentions, Message, MessageFlags, MessageReference
    from .users import User


class ChannelType(enum.IntFlag):
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_ANNOUNCEMENT = 5
    ANNOUNCEMENT_THREAD = 10
    PUBLIC_THREAD = 11
    PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13
    GUILD_DIRECTORY = 14
    GUILD_FORUM = 15


@attrs.define(kw_only=True, slots=True, repr=True)
class ChannelLike(DiscordObject):
    client: GatewayClient
    id: Snowflake
    """ID of the channel-like object."""
    raw: dict[str, typing.Any]
    """The raw payload for this channel."""
    type: ChannelType
    """Type of the channel."""


class ImplementsMessages(abc.ABC):
    id: Snowflake
    client: GatewayClient
    last_message_id: Snowflake | None

    async def fetch_message(self, message_id: int, /) -> Message:
        return await self.client.rest.fetch_message(self.id, message_id)

    async def delete_message(self, message_id: int, /) -> None:
        return await self.client.rest.delete_message(self.id, message_id)

    async def send(
        self,
        content: str | None = None,
        *,
        embed: EmbedConstructor | None = None,
        embeds: typing.Sequence["EmbedConstructor"] = (),
        components: typing.Sequence[ActionRowContainer] = (),
        reference: int | MessageReference | None = None,
        allowed_mentions: AllowedMentions | None = None,
        flags: MessageFlags | None = None,
        delete_after: int | None = None,
    ) -> Message:
        if embed:
            embeds = [embed]

        return await self.client.rest.create_message(
            self.id,
            content,
            embeds=embeds,
            components=components,
            reference=reference,
            allowed_mentions=allowed_mentions,
            delete_after=delete_after,
            flags=flags,
        )


class ImplementsVoice(abc.ABC):
    id: Snowflake
    guild_id: Snowflake
    client: GatewayClient

    async def connect(self, self_deaf: bool = False, self_mute: bool = False) -> None:
        await self.client.gateway.update_voice_state(self.guild_id, self.id, self_mute=self_mute, self_deaf=self_deaf)


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildChannel(ChannelLike):
    name: str
    """Name of the channel."""
    guild_id: Snowflake
    """ID of the guild this channel belongs to."""
    position: int
    """Position of the channel."""
    permissions_overwrites: list[PermissionOverwrites]
    """List of Permission overwrites in this channel."""


@attrs.define(kw_only=True, slots=True, repr=True)
class DMChannel(ImplementsMessages, ChannelLike):
    type: ChannelType = attrs.field(init=False, default=ChannelType.DM)
    last_message_id: Snowflake | None
    """ID of the last message sent in the channel."""
    recipients: list[User]

    @property
    def user(self) -> User | None:
        return self.recipients[0] if self.recipients else None


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildTextChannel(ImplementsMessages, GuildChannel):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_TEXT)
    rate_limit_per_user: int
    """The cooldown in the channel per-user."""
    nsfw: bool
    """True if this channel is NSFW marked."""
    topic: str
    """Topic of the channel."""
    last_message_id: Snowflake | None
    """ID of the last message sent in the channel."""
    parent_id: Snowflake | None
    """ID of the category channel this channel belongs to. """
    default_auto_archive_duration: int
    """Time after which threads get archived by default."""


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildVoiceChannel(GuildChannel, ImplementsMessages, ImplementsVoice):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_VOICE)

    parent_id: Snowflake | None
    """ID of the category channel this channel belongs to. """
    bitrate: int
    """Bitrate for the channel."""
    user_limit: int
    """Max number of users allowed in the channel."""
    rtc_region: str | None
    """The rtc region."""
    rate_limit_per_user: int
    """The cooldown in the channel per-user."""
    nsfw: bool
    """True if the channel is marked NSFW"""


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildCategoryChannel(GuildChannel):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_CATEGORY)


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildAnnouncmentChannel(GuildChannel, ImplementsMessages):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_ANNOUNCEMENT)
    nsfw: bool
    """True if the channel is NSFW marked."""
    topic: str
    """Topic of the channel."""
    last_message_id: Snowflake | None
    """ID of the last message sent in the channel."""
    parent_id: Snowflake | None
    """ID of the category channel this channel belongs to. """
    default_auto_archive_duration: int
    """Time after which threads get archived by default."""


@attrs.define(kw_only=True, slots=True, repr=True)
class ThreadMetaData:
    archived: bool
    auto_archive_duration: int
    archive_timestamp: datetime.datetime | None
    locked: bool


@attrs.define(kw_only=True, slots=True, repr=True)
class ThreadChannel(GuildChannel, ImplementsMessages):
    type: ChannelType
    parent_id: Snowflake
    """ID of the parent channel."""
    owner_id: Snowflake
    """ID of the thread creator."""
    last_message_id: Snowflake | None
    """ID of last message created in the channel."""
    message_count: int
    """Number of messages in the thread."""
    member_count: int
    rate_limit_per_user: int
    total_message_sent: int

    @property
    def is_private(self) -> bool:
        return True if self.type is ChannelType.PRIVATE_THREAD else False


@attrs.define(kw_only=True, slots=True, repr=True)
class ThreadAnnouncementChannel(ThreadChannel):
    type: ChannelType = attrs.field(init=False, default=ChannelType.ANNOUNCEMENT_THREAD)


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildStageVoiceChannel(GuildChannel, ImplementsVoice):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_STAGE_VOICE)
    nsfw: bool


class GuildDirectoryChannel:  # not implemented by discord yet.
    ...


@attrs.define(kw_only=True, slots=True, repr=True)
class GuildForumChannel(GuildChannel):
    type: ChannelType = attrs.field(init=False, default=ChannelType.GUILD_FORUM)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import abc
import typing

import attrs

if typing.TYPE_CHECKING:
    from wyvern._types import AutocompleteT


__all__: tuple[str, ...] = ("SlashOptionConverter",)


@attrs.define(kw_only=True)
class BaseCallable(abc.ABC):
    callback: typing.Any

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return self.callback(*args, **kwargs)


class HasAutoComplete(abc.ABC):
    autocompletes: dict[str, "AutocompleteT"] = {}

    async def with_autocomplete(
        self, opt_name: str | None = None, *, opt_names: typing.Sequence[str] = ()
    ) -> typing.Callable[..., "AutocompleteT"]:
        def inner(callback: "AutocompleteT") -> "AutocompleteT":
            nonlocal opt_name, opt_names
            if opt_name is not None and opt_names == ():
                self.autocompletes[opt_name] = callback
            elif opt_names != ():
                for opt in opt_names:
                    self.autocompletes[opt] = callback
            else:
                raise ValueError("You need to provide one of the `opt_name` or `opt_names` values.")

            return callback

        return inner


class HasSubCommands(abc.ABC):
    subcommands: dict[str, typing.Any] = {}

    @abc.abstractmethod
    def with_subcommand(self, *, name: str, description: str) -> typing.Any:
        ...


class SlashOptionConverter(abc.ABC):
    async def convert(self, arg: str) -> typing.Any:
        ...



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.events import Event, EventListener, as_listener

__all__: tuple[str, ...] = ("Plugin",)


class Plugin:
    """Plugins for extendibility.

    Parameters
    ----------
    name: str
        Name of the plugin.
    description : str | None
        Description of the plugin.
    """

    name: str
    """Name of the plugin."""
    description: str | None = None
    """Description of the plugin."""
    client: GatewayClient
    """Client attached to the plugin."""
    listeners: list[EventListener] = []
    """EventListeners attached to the plugin."""

    def __init__(self, name: str, *, description: str | None = None) -> None:
        self.name = name
        self.description = description

    def setup_plugin(self, client: "GatewayClient") -> None:
        self.client = client
        for listener in self.listeners:
            client.event_handler.add_listener(listener)

    def with_listener(
        self, event: Event, *, max_trigger: int | float = float("inf")
    ) -> typing.Callable[[typing.Callable[..., typing.Awaitable[typing.Any]]], EventListener]:
        """
        Creates and adds a new listenet to the client's event handler.

        Parameters
        ----------
        event:  wyvern.events.Event
            The event to listen.
        max_trigger: int | float
            Maximum number of times this listener has to be triggered.

        Returns
        -------
        wyvern.events.EventListener
            A EventListener object.
        """

        def inner(callback: typing.Callable[..., typing.Awaitable[typing.Any]]) -> EventListener:
            lsnr = as_listener(event, max_trigger=max_trigger)(callback)
            self.listeners.append(lsnr)
            return lsnr

        return inner



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import time
import typing

from .enums import WSEventEnums

if typing.TYPE_CHECKING:
    from wyvern.gateway import Gateway

__all__: tuple[str, ...] = ("KeepAlive",)


class KeepAlive:
    sequence: int = 0
    last_heartbeat: float

    async def start(self, ws: "Gateway") -> None:
        while True:
            await ws.socket.send_json({"op": WSEventEnums.HEARTBEAT, "d": self.sequence})
            self.last_heartbeat = time.perf_counter()
            await asyncio.sleep(ws._heartbeat_interval)



from __future__ import annotations

import datetime

import attrs

from wyvern.files import Avatar, AvatarType

from .base import Snowflake
from .users import User


@attrs.define(kw_only=True, slots=True, repr=True)
class Member(User):
    user: User
    """User of the member."""
    guild_id: Snowflake
    """ID of the guild this member belongs to."""
    nickname: str | None
    """Nickname of the member, if any."""
    role_ids: list[Snowflake]
    """List of role ids."""
    joined_at: datetime.datetime
    """Datetime on which the member joined the guild."""
    premium_since: datetime.datetime | None
    """Datetime since member has been boosting the guild."""
    deaf: bool | None
    """True if member is deafened."""
    mute: bool | None
    """True if member is muted."""
    pending: bool
    """True if member is in pending state."""
    communication_disabled_until: datetime.datetime | None
    """Timeout for the member."""
    guild_avatar_hash: str | None
    """Avatar hash for the guild."""

    def update_state(self) -> "Member" | None:
        return member if (member := self.client.members.get(self.guild_id, self.id)) else None

    @property
    def display_avatar(self) -> Avatar:
        """
        Returns
        -------
        str
            Member's visible avatar.
        """
        return self.guild_avatar or self.avatar or self.default_avatar

    @property
    def display_name(self) -> str:
        """
        Returns
        -------
        str
            Member's visible name.
        """
        return self.nickname or self.username

    @property
    def guild_avatar(self) -> Avatar | None:
        """
        Returns
        -------
        None | wyvern.Avatar
            The user's guild avatar.
        """
        return (
            Avatar(client=self.client, type=AvatarType.GUILD, hash=self.guild_avatar_hash)
            if self.guild_avatar_hash
            else None
        )



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import annotations

import typing

import attrs

if typing.TYPE_CHECKING:
    from wyvern.rest import RequestRoute


class WyvernException(Exception):
    ...


@attrs.define
class HTTPException(WyvernException):
    message: str
    code: int
    route: "RequestRoute"

    def __str__(self) -> str:
        return self.message

    def create(self) -> HTTPException:
        return get_http_exception(self.code)(self.message, self.code, self.route)


class BadRequest(HTTPException):
    ...


class Unauthorized(HTTPException):
    ...


class Forbidden(HTTPException):
    ...


class NotFound(HTTPException):
    ...


class InvalidMethod(HTTPException):
    ...


class Ratelimited(HTTPException):
    ...


class UnknownError(HTTPException):
    ...


class UserNotFound(NotFound):
    """Raised when the targetted user was not found."""


class MemberNotFound(NotFound):
    """Raised when the targetted member was not found."""


excs: dict[int, typing.Type[HTTPException]] = {
    400: BadRequest,
    401: Unauthorized,
    403: Forbidden,
    404: NotFound,
    405: InvalidMethod,
    429: Ratelimited,
}


def get_http_exception(code: int) -> typing.Type[HTTPException]:
    return excs.get(code, UnknownError)


class CommandAlreadyExists(WyvernException):
    ...


class PluginException(WyvernException):
    ...



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Utility functions and classes for the library."""
from __future__ import annotations

import ast
import datetime
import inspect
import typing

import attrs

__all__: tuple[str, ...] = ("Empty", "EMPTY", "get_arg_count", "create_timestamp", "Eval", "Hook", "as_hook")


class Empty:
    ...


EMPTY = Empty()


def get_arg_count(callable: typing.Callable[..., typing.Any]) -> int:
    """Counts the number of args in a callable.

    Parameters
    ----------
    callable : typing.Callable[..., typing.Any]
        The callable to get arg count of.

    Returns
    -------
    int
        Number of arguments in a callable.
    """
    return len(inspect.getargs(callable.__code__).args)


def copy_docs(_from: typing.Any, replace: dict[str, str] = {}) -> typing.Callable[[typing.Any], typing.Any]:
    def copy(_to: typing.Any) -> typing.Any:
        _to.__doc__ = _from.__doc__
        [_to.__doc__.replace(text, replacement) for text, replacement in replace.items()]
        return _to

    return copy


def create_timestamp(dt: datetime.datetime | datetime.timedelta, *, style: str = "t") -> str:
    """Creates an UNIX timestamp for provided datetime or timedelta object.

    Parameters
    ----------
    dt : datetime.datetime | datetime.timedelta
        The datetime or timedelta to convert.

    Returns
    -------
    str
        The UNIX timestamp.
    """
    if isinstance(dt, datetime.timedelta):
        dt = datetime.datetime.utcnow() + dt
    return f"<t:{int(dt.timestamp())}:{style}>"


@attrs.define(kw_only=True)
class CacheConfigurations:
    users: bool = True
    guilds: bool = True
    members: bool = True
    roles: bool = True
    channels: bool = True
    messages: bool = False


@attrs.define
class Hook:
    """Hooks for per-module setup. They can be loaded using `GatewayClient.load_hooks`"""

    callback: typing.Callable[..., typing.Any]
    """Callback of the hook."""
    name: str
    """Name of the hook."""

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return self.callback(*args, **kwargs)


def as_hook(name: str | None = None) -> typing.Callable[..., Hook]:
    """Creates a [wyvern.utils.Hook][]

    Parameters
    ----------
    name : str
        Name of the hook.

    Returns
    -------
    wyvern.utils.Hook
        The hook that was created.
    """

    def inner(callback: typing.Callable[..., typing.Any]) -> Hook:
        return Hook(callback, name or callback.__name__)

    return inner


class Eval:
    """Class for code evaluation.

    !!! warning
        This class is not sandboxed so data like environmental variables
        will be evaluated when the methods gets executed as well.

    """

    def add_returns(self, body: typing.Any) -> None:
        if isinstance(body[-1], ast.Expr):
            body[-1] = ast.Return(body[-1].value)
            ast.fix_missing_locations(body[-1])

        # for if statements, we insert returns into the body and the orelse
        if isinstance(body[-1], ast.If):
            self.add_returns(body[-1].body)
            self.add_returns(body[-1].orelse)

        # for with blocks, again we insert returns into the body
        if isinstance(body[-1], ast.With):
            self.add_returns(body[-1].body)

    async def f_eval(self, *, code: str, renv: dict[str, typing.Any]) -> typing.Any:
        """Evaluates the code in the bot's namespace.

        Parameters
        ----------
        code : str
            The code to evaluate.
        renv: dict[str, typing.Any]
            Environment to evaluate code in.

        Returns
        -------
        typing.Any
            The result of the code.

        """
        _fn_name = "__wyvern_eval"
        code = "\n".join(f"    {i}" for i in code.strip().splitlines())
        parsed: typing.Any = ast.parse(f"async def {_fn_name}:\n{code}")
        self.add_returns(parsed.body[0].body)
        exec(compile(parsed, filename="<ast>", mode="exec"), renv)
        fn = renv[_fn_name]
        return await fn()



from . import interactions
from .colors import Color

AppCommandInter = interactions.ApplicationCommandInteraction
"""Alias for ApplicationCommandInteraction"""
Colour = Color
"""Alias for Color class as per British english"""



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.models.members import Member

__all__: tuple[str, ...] = ("MembersState",)


class MembersState:
    _client: "GatewayClient"
    cached_members: dict[int, dict[int, Member]] = {}

    def __init__(self, client: GatewayClient) -> None:
        self._client = client

    def _guild_check(self, gid: int) -> None:
        if gid not in self.cached_members.keys():
            self.cached_members[gid] = {}

    def get(self, guild_id: int, member_id: int) -> "Member" | None:
        self._guild_check(guild_id)
        return self.cached_members[guild_id].get(member_id)

    async def fetch(self, guild_id: int, member_id: int) -> "Member":
        return await self._client.rest.fetch_member(guild_id, member_id)

    def get_member_named(self, guild_id: int, name: str) -> "Member" | None:
        self._guild_check(guild_id)
        return (
            members[0]
            if (members := [member for member in self.cached_members[guild_id].values() if member.username == name])
            else None
        )

    def add_member(self, mem: Member) -> None:
        self._guild_check(mem.guild_id)
        self.cached_members[mem.guild_id][mem.id] = mem

    def update_user_state(self) -> None:
        [
            self._client.users._add_to_state(*list(map(lambda mem: mem.user, members.values())))
            for members in self.cached_members.values()
        ]



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import inspect
import typing

if typing.TYPE_CHECKING:
    from wyvern.interactions.applications import ApplicationCommandInteraction
    from wyvern.models.messages import Message

    from .slash_commands import SlashCommand

__all__: tuple[str, ...] = ("CommandHandler",)


class CommandHandler:
    """Command Handler for interacting with interaction and prefix commands.

    !!! note
        This class is purely created to be implemented with the GatewayClient subclass
        for the [wyvern.CommandsClient][] class.
    """

    prefix_getter: typing.Any = None
    prefix_type: type | None = None
    slash_commands: dict[str, SlashCommand]

    async def get_prefixes(self, message: Message) -> typing.Any:
        """Get's a list of prefix with respect to the message.

        Parameters
        ----------

        message : wyvern.Message
            The message to get prefixes for.

        Returns
        -------

        typing.Any
        """
        if self.prefix_type is str:
            return [self.prefix_getter]
        elif type(self.prefix_getter) in [set, list, tuple]:
            return self.prefix_getter
        else:
            if inspect.iscoroutinefunction(self.prefix_getter):
                return [prefix] if isinstance((prefix := await self.prefix_getter(self, message)), str) else prefix
            elif inspect.isfunction(self.prefix_getter):
                return [prefix] if isinstance((prefix := self.prefix_getter(self, message)), str) else prefix

    async def process_application_commands(self, interaction: ApplicationCommandInteraction) -> None:
        """Processes application command interactions for app commands.

        Parameters
        ----------

        interaction : wyvern.ApplicationCommandInteraction
            The interaction to process.
        """
        command = self.slash_commands.get(interaction.data.command_name)
        if command is not None:
            await command(
                interaction,
            )
